@@ -11,6 +11,7 @@
 
 #include "hdy-animation.h"
 #include "hdy-action-row.h"
+#include "hdy-deck.h"
 #include "hdy-preferences-group-private.h"
 #include "hdy-preferences-page-private.h"
 #include "hdy-view-switcher.h"
@@ -25,11 +26,18 @@
  * The #HdyPreferencesWindow widget presents an application's preferences
  * gathered into pages and groups. The preferences are searchable by the user.
  *
+ * # CSS nodes
+ *
+ * #HdyPreferencesWindow has a main CSS node with the name window and the style
+ * class .preferences.
+ *
  * Since: 0.0.10
  */
 
 typedef struct
 {
+  HdyDeck *subpages_deck;
+  GtkWidget *preferences;
   GtkStack *content_stack;
   GtkStack *pages_stack;
   GtkToggleButton *search_button;
@@ -41,7 +49,9 @@ typedef struct
   HdyViewSwitcherTitle *view_switcher_title;
 
   gboolean search_enabled;
+  gboolean can_swipe_back;
   gint n_last_search_results;
+  GtkWidget *subpage;
 } HdyPreferencesWindowPrivate;
 
 G_DEFINE_TYPE_WITH_PRIVATE (HdyPreferencesWindow, hdy_preferences_window, HDY_TYPE_WINDOW)
@@ -49,18 +59,61 @@ G_DEFINE_TYPE_WITH_PRIVATE (HdyPreferencesWindow, hdy_preferences_window, HDY_TY
 enum {
   PROP_0,
   PROP_SEARCH_ENABLED,
+  PROP_CAN_SWIPE_BACK,
   LAST_PROP,
 };
 
 static GParamSpec *props[LAST_PROP];
 
+/* Copied and modified from gtklabel.c, separate_uline_pattern() */
+static gchar *
+strip_mnemonic (const gchar *src)
+{
+  g_autofree gchar *new_str = g_new (gchar, strlen (src) + 1);
+  gchar *dest = new_str;
+  gboolean underscore = FALSE;
+
+  while (*src) {
+    gunichar c;
+    const gchar *next_src;
+
+    c = g_utf8_get_char (src);
+    if (c == (gunichar) -1) {
+      g_warning ("Invalid input string");
+
+      return NULL;
+    }
+
+    next_src = g_utf8_next_char (src);
+
+    if (underscore) {
+      while (src < next_src)
+        *dest++ = *src++;
+
+      underscore = FALSE;
+    } else {
+      if (c == '_'){
+        underscore = TRUE;
+        src = next_src;
+      } else {
+        while (src < next_src)
+          *dest++ = *src++;
+      }
+    }
+  }
+
+  *dest = 0;
+
+  return g_steal_pointer (&new_str);
+}
+
 static gboolean
 filter_search_results (HdyActionRow         *row,
                        HdyPreferencesWindow *self)
 {
   HdyPreferencesWindowPrivate *priv = hdy_preferences_window_get_instance_private (self);
   g_autofree gchar *text = g_utf8_casefold (gtk_entry_get_text (GTK_ENTRY (priv->search_entry)), -1);
-  g_autofree gchar *title = g_utf8_casefold (hdy_action_row_get_title (row), -1);
+  g_autofree gchar *title = g_utf8_casefold (hdy_preferences_row_get_title (HDY_PREFERENCES_ROW (row)), -1);
   g_autofree gchar *subtitle = NULL;
 
   /* The CSS engine works in such a way that invisible children are treated as
@@ -74,7 +127,16 @@ filter_search_results (HdyActionRow         *row,
    * See https://gitlab.gnome.org/GNOME/libhandy/-/merge_requests/424
    */
 
-  if (strstr (title, text)) {
+  if (hdy_preferences_row_get_use_underline (HDY_PREFERENCES_ROW (row))) {
+    gchar *stripped_title = strip_mnemonic (title);
+
+    if (stripped_title) {
+      g_free (title);
+      title = stripped_title;
+    }
+  }
+
+  if (!!strstr (title, text)) {
     priv->n_last_search_results++;
     gtk_widget_show (GTK_WIDGET (row));
 
@@ -166,9 +228,6 @@ search_result_activated_cb (HdyPreferencesWindow *self,
   HdyPreferencesWindowPrivate *priv = hdy_preferences_window_get_instance_private (self);
   HdyPreferencesPage *page;
   HdyPreferencesRow *row;
-  GtkAdjustment *adjustment;
-  GtkAllocation allocation;
-  gint y = 0;
 
   gtk_toggle_button_set_active (priv->search_button, FALSE);
   page = HDY_PREFERENCES_PAGE (g_object_get_data (G_OBJECT (widget), "page"));
@@ -177,21 +236,9 @@ search_result_activated_cb (HdyPreferencesWindow *self,
   g_assert (page != NULL);
   g_assert (row != NULL);
 
-  adjustment = hdy_preferences_page_get_vadjustment (page);
-
-  g_assert (adjustment != NULL);
-
   gtk_stack_set_visible_child (priv->pages_stack, GTK_WIDGET (page));
   gtk_widget_set_can_focus (GTK_WIDGET (row), TRUE);
   gtk_widget_grab_focus (GTK_WIDGET (row));
-
-  if (!gtk_widget_translate_coordinates (GTK_WIDGET (row), GTK_WIDGET (page), 0, 0, NULL, &y))
-    return;
-
-  gtk_container_set_focus_child (GTK_CONTAINER (page), GTK_WIDGET (row));
-  y += gtk_adjustment_get_value (adjustment);
-  gtk_widget_get_allocation (GTK_WIDGET (row), &allocation);
-  gtk_adjustment_clamp_page (adjustment, y, y + allocation.height);
 }
 
 static gboolean
@@ -203,10 +250,30 @@ key_press_event_cb (GtkWidget            *sender,
   GdkModifierType default_modifiers = gtk_accelerator_get_default_mod_mask ();
   guint keyval;
   GdkModifierType state;
+  GdkKeymap *keymap;
+  GdkEventKey *key_event = (GdkEventKey *) event;
 
-  gdk_event_get_keyval (event, &keyval);
   gdk_event_get_state (event, &state);
 
+  keymap = gdk_keymap_get_for_display (gtk_widget_get_display (sender));
+
+  gdk_keymap_translate_keyboard_state (keymap,
+                                       key_event->hardware_keycode,
+                                       state,
+                                       key_event->group,
+                                       &keyval, NULL, NULL, NULL);
+
+  if (priv->subpage) {
+    if (keyval == GDK_KEY_Escape &&
+        hdy_preferences_window_get_can_swipe_back (self)) {
+      hdy_preferences_window_close_subpage (self);
+
+      return GDK_EVENT_STOP;
+    }
+
+    return GDK_EVENT_PROPAGATE;
+  }
+
   if (priv->search_enabled &&
       (keyval == GDK_KEY_f || keyval == GDK_KEY_F) &&
       (state & default_modifiers) == GDK_CONTROL_MASK) {
@@ -222,9 +289,45 @@ key_press_event_cb (GtkWidget            *sender,
     return GDK_EVENT_STOP;
   }
 
+  if (keyval == GDK_KEY_Escape) {
+    gtk_window_close (GTK_WINDOW (self));
+
+    return GDK_EVENT_STOP;
+  }
+
   return GDK_EVENT_PROPAGATE;
 }
 
+static void
+try_remove_subpages (HdyPreferencesWindow *self)
+{
+  HdyPreferencesWindowPrivate *priv = hdy_preferences_window_get_instance_private (self);
+
+  if (hdy_deck_get_transition_running (priv->subpages_deck))
+    return;
+
+  if (hdy_deck_get_visible_child (priv->subpages_deck) == priv->preferences)
+    priv->subpage = NULL;
+
+  for (GList *child = gtk_container_get_children (GTK_CONTAINER (priv->subpages_deck));
+       child;
+       child = child->next)
+    if (child->data != priv->preferences && child->data != priv->subpage)
+      gtk_container_remove (GTK_CONTAINER (priv->subpages_deck), child->data);
+}
+
+static void
+subpages_deck_transition_running_cb (HdyPreferencesWindow *self)
+{
+  try_remove_subpages (self);
+}
+
+static void
+subpages_deck_visible_child_cb (HdyPreferencesWindow *self)
+{
+  try_remove_subpages (self);
+}
+
 static void
 header_bar_size_allocate_cb (HdyPreferencesWindow *self,
                              GdkRectangle         *allocation)
@@ -336,6 +439,9 @@ hdy_preferences_window_get_property (GObject    *object,
   case PROP_SEARCH_ENABLED:
     g_value_set_boolean (value, hdy_preferences_window_get_search_enabled (self));
     break;
+  case PROP_CAN_SWIPE_BACK:
+    g_value_set_boolean (value, hdy_preferences_window_get_can_swipe_back (self));
+    break;
   default:
     G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
   }
@@ -353,6 +459,9 @@ hdy_preferences_window_set_property (GObject      *object,
   case PROP_SEARCH_ENABLED:
     hdy_preferences_window_set_search_enabled (self, g_value_get_boolean (value));
     break;
+  case PROP_CAN_SWIPE_BACK:
+    hdy_preferences_window_set_can_swipe_back (self, g_value_get_boolean (value));
+    break;
   default:
     G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
   }
@@ -440,10 +549,26 @@ hdy_preferences_window_class_init (HdyPreferencesWindowClass *klass)
                           TRUE,
                           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
+  /**
+   * HdyPreferencesWindow:can-swipe-back:
+   *
+   * Whether or not the window allows closing the subpage via a swipe gesture.
+   *
+   * Since: 1.0
+   */
+  props[PROP_CAN_SWIPE_BACK] =
+      g_param_spec_boolean ("can-swipe-back",
+                            _("Can swipe back"),
+                            _("Whether or not swipe gesture can be used to switch from a subpage to the preferences"),
+                            FALSE,
+                            G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY);
+
   g_object_class_install_properties (object_class, LAST_PROP, props);
 
   gtk_widget_class_set_template_from_resource (widget_class,
                                                "/sm/puri/handy/ui/hdy-preferences-window.ui");
+  gtk_widget_class_bind_template_child_private (widget_class, HdyPreferencesWindow, subpages_deck);
+  gtk_widget_class_bind_template_child_private (widget_class, HdyPreferencesWindow, preferences);
   gtk_widget_class_bind_template_child_private (widget_class, HdyPreferencesWindow, content_stack);
   gtk_widget_class_bind_template_child_private (widget_class, HdyPreferencesWindow, pages_stack);
   gtk_widget_class_bind_template_child_private (widget_class, HdyPreferencesWindow, search_button);
@@ -453,6 +578,8 @@ hdy_preferences_window_class_init (HdyPreferencesWindowClass *klass)
   gtk_widget_class_bind_template_child_private (widget_class, HdyPreferencesWindow, title_stack);
   gtk_widget_class_bind_template_child_private (widget_class, HdyPreferencesWindow, view_switcher_bar);
   gtk_widget_class_bind_template_child_private (widget_class, HdyPreferencesWindow, view_switcher_title);
+  gtk_widget_class_bind_template_callback (widget_class, subpages_deck_transition_running_cb);
+  gtk_widget_class_bind_template_callback (widget_class, subpages_deck_visible_child_cb);
   gtk_widget_class_bind_template_callback (widget_class, header_bar_size_allocate_cb);
   gtk_widget_class_bind_template_callback (widget_class, title_stack_notify_transition_running_cb);
   gtk_widget_class_bind_template_callback (widget_class, title_stack_notify_visible_child_cb);
@@ -543,3 +670,116 @@ hdy_preferences_window_set_search_enabled (HdyPreferencesWindow *self,
 
   g_object_notify_by_pspec (G_OBJECT (self), props[PROP_SEARCH_ENABLED]);
 }
+
+/**
+ * hdy_preferences_window_set_can_swipe_back:
+ * @self: a #HdyPreferencesWindow
+ * @can_swipe_back: the new value
+ *
+ * Sets whether or not @self allows switching from a subpage to the preferences
+ * via a swipe gesture.
+ *
+ * Since: 1.0
+ */
+void
+hdy_preferences_window_set_can_swipe_back (HdyPreferencesWindow *self,
+                                           gboolean              can_swipe_back)
+{
+  HdyPreferencesWindowPrivate *priv;
+
+  g_return_if_fail (HDY_IS_PREFERENCES_WINDOW (self));
+
+  priv = hdy_preferences_window_get_instance_private (self);
+
+  can_swipe_back = !!can_swipe_back;
+
+  if (priv->can_swipe_back == can_swipe_back)
+    return;
+
+  priv->can_swipe_back = can_swipe_back;
+
+  g_object_notify_by_pspec (G_OBJECT (self), props[PROP_CAN_SWIPE_BACK]);
+}
+
+/**
+ * hdy_preferences_window_get_can_swipe_back
+ * @self: a #HdyPreferencesWindow
+ *
+ * Returns whether or not @self allows switching from a subpage to the
+ * preferences via a swipe gesture.
+ *
+ * Returns: %TRUE if back swipe is enabled.
+ *
+ * Since: 1.0
+ */
+gboolean
+hdy_preferences_window_get_can_swipe_back (HdyPreferencesWindow *self)
+{
+  HdyPreferencesWindowPrivate *priv;
+
+  g_return_val_if_fail (HDY_IS_PREFERENCES_WINDOW (self), FALSE);
+
+  priv = hdy_preferences_window_get_instance_private (self);
+
+  return priv->can_swipe_back;
+}
+
+/**
+ * hdy_preferences_window_present_subpage:
+ * @self: a #HdyPreferencesWindow
+ * @subpage: the subpage
+ *
+ * Sets @subpage as the window's subpage and present it.
+ * The transition can be cancelled by the user, in which case visible child will
+ * change back to the previously visible child.
+ *
+ * Since: 1.0
+ */
+void
+hdy_preferences_window_present_subpage (HdyPreferencesWindow *self,
+                                        GtkWidget            *subpage)
+{
+  HdyPreferencesWindowPrivate *priv;
+
+  g_return_if_fail (HDY_IS_PREFERENCES_WINDOW (self));
+  g_return_if_fail (GTK_IS_WIDGET (subpage));
+
+  priv = hdy_preferences_window_get_instance_private (self);
+
+  if (priv->subpage == subpage)
+    return;
+
+  priv->subpage = subpage;
+
+  /* The check below avoids a warning when re-entering a subpage during the
+   * transition between the that subpage to the preferences.
+   */
+  if (gtk_widget_get_parent (subpage) != GTK_WIDGET (priv->subpages_deck))
+    gtk_container_add (GTK_CONTAINER (priv->subpages_deck), subpage);
+
+  hdy_deck_set_visible_child (priv->subpages_deck, subpage);
+}
+
+/**
+ * hdy_preferences_window_close_subpage:
+ * @self: a #HdyPreferencesWindow
+ *
+ * Closes the current subpage to return back to the preferences, if there is no
+ * presented subpage, this does nothing.
+ *
+ * Since: 1.0
+ */
+void
+hdy_preferences_window_close_subpage (HdyPreferencesWindow *self)
+{
+  HdyPreferencesWindowPrivate *priv;
+
+  g_return_if_fail (HDY_IS_PREFERENCES_WINDOW (self));
+
+  priv = hdy_preferences_window_get_instance_private (self);
+
+  if (priv->subpage == NULL)
+    return;
+
+  hdy_deck_set_visible_child (priv->subpages_deck, priv->preferences);
+}
